from __future__ import annotations
import numpy as np
import sys
from . import jacktools
import signal
import math
import textwrap
from typing import TYPE_CHECKING
import emlib.dialogs
import subprocess

if TYPE_CHECKING:
    from .instr import Instr
    from typing import *
    from csoundlib import AudioDevice
    

_registry: Dict[str, Any] = {}


def isrunning(prog: str) -> bool:
    "True if prog is running"
    failed = subprocess.call(['pgrep', '-f', prog],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return not failed


def m2f(midinote: float, a4:float) -> float:
    """
    Convert a midi-note to a frequency
    """
    return 2**((midinote-69)/12.0)*a4


def arrayNumChannels(a: np.ndarray) -> int:
    """
    Return the number of channels in a numpy array holding audio data
    """
    return 1 if len(a.shape) == 1 else a.shape[1]


def getChannel(samples: np.ndarray, channel: int) -> np.ndarray:
    """ Get a channel of a numpy array holding possibly multichannel
    audio data

    Args:
        samples: the (multichannel) audio data
        channel: the index of the channel to extract

    Returns:
        a numpy array holding a channel of audio data.
    """
    return samples if len(samples.shape) == 1 else samples[:, channel]


def sigintHandler(sig, frame):
    print(frame)
    raise KeyboardInterrupt("SIGINT (CTRL-C) while waiting")


def setSigintHandler():
    """
    Set own sigint handler to prevent CTRL-C from crashing csound

    It will do nothing if this was already set
    """
    if _registry.get('sigint_handler_set'):
        return
    original_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, sigintHandler)
    _registry['original_sigint_handler'] = original_handler
    _registry['sigint_handler_set'] = True


def removeSigintHandler():
    """
    Reset the sigint handler to its original state
    This will do nothing if our own handler was not set
    in the first place
    """
    if not _registry.get('sigint_handler_set'):
        return
    signal.signal(signal.SIGINT, _registry['original_sigint_handler'])
    _registry['sigint_handler_set'] = False


def determineNumbuffers(backend:str, buffersize:int) -> int:
    if backend == 'jack':
        info = jacktools.getInfo()
        numbuffers = int(math.ceil(info.blocksize / buffersize))
    else:
        numbuffers = 2
    return numbuffers


def instrResolveArgs(instr: Instr,
                     p4: int,
                     pargs: Union[List[float], Dict[str, float]]=None,
                     pkws: Dict[str, float]=None
                     ) -> List[float]:
    allargs: List[float] = [float(p4)]
    if not pargs and not instr.pargsIndexToDefaultValue and not pkws:
        return allargs
    if isinstance(pargs, list):
        allargs.extend(instr.pargsTranslate(pargs, pkws))
    else:
        if pkws:
            if pargs:
                pargs.update(pkws)
            else:
                pargs = pkws
        allargs.extend(instr.pargsTranslate(kws=pargs))
    return allargs


def instrWrapBody(body:str, instrid:Union[int, str, Sequence[str]], comment:str= '',
                  addNotificationCode=False) -> str:
    s = r"""
instr {instrnum}  {commentstr}
    {notifystr}
    {body}
endin
    """
    if addNotificationCode:
        # notifystr = 'defer "outvalue", "__dealloc__", p1'
        # TODO: defer can cause memory corruption in some cases
        notifystr = 'atstop "_notifyDealloc", 0.01, 0.0, p1'
    else:
        notifystr = ''
    commentstr = "; " + comment if comment else ""
    if isinstance(instrid, (list, tuple)):
        instrid = ", ".join([str(i) for i in instrid])
    s = s.format(instrnum=instrid, body=body, notifystr=notifystr,
                 commentstr=commentstr)
    s = textwrap.dedent(s)
    return s


def addLineNumbers(code: str) -> str:
    lines = [f"{i:03d}  {line}"
             for i, line in enumerate(code.splitlines(), start=1)]
    return "\n".join(lines)


# Maps platform values as given by sys.platform to more readable aliases
_platformAliases = {
    'linux2': 'linux',
    'linux': 'linux',
    'darwin': 'macos',
    'macos': 'macos',
    'win32': 'windows',
    'windows': 'windows'
}

platform = _platformAliases[sys.platform]


# Maps possible platform names to names as returned by sys.platform
_normalizedPlatforms = {
    'linux': 'linux',
    'win32': 'win32',
    'darwin': 'darwin',
    'windows': 'win32',
    'macos': 'darwin'
}

def platformAlias(platform: str) -> str:
    """
    Return the platform alias (macos, windows, linux) for the
    given platform (instead of darwin, win32, etc)

    This is the opposite of `normalizePlatform`
    """
    out = _platformAliases.get(platform)
    if out is None:
        raise KeyError(f"Platform {platform} unknown, possible values are"
                       f" {_platformAliases.keys()}")
    return out


def normalizePlatform(s:str) -> str:
    """Return the platform as given by sys.platform

    This is the opposite of `platformAlias`
    """
    out = _normalizedPlatforms.get(s)
    if out is None:
        raise KeyError(f"Platform {s} not known")
    return out


def resolveOption(prioritizedOptions:List[str], availableOptions:List[str]
                  ) -> Optional[str]:
    for opt in prioritizedOptions:
        if opt in availableOptions:
            return opt
    return None


def selectAudioDevice(devices: List[AudioDevice], title='Select device'
                      ) -> Optional[AudioDevice]:
    if len(devices) == 1:
        return devices[0]
    outnames = [dev.info() for dev in devices]
    selected = emlib.dialogs.selectItem(items=outnames, title=title)
    if not selected:
        return None
    idx = outnames.index(selected)
    outdev = devices[idx]
    return outdev


def selectItem(items: List[str], title="Select") -> Optional[str]:
    return emlib.dialogs.selectItem(items=items, title=title)


def instrNameFromP1(p1: Union[float, str]) -> Union[int, str]:
    return int(p1) if isinstance(p1, (int, float)) else p1.split(".")[0]


def resolvePfieldIndex(pfield: Union[int, str], pfieldNameToIndex: Dict[str, int] = None
                       ) -> int:
    if isinstance(pfield, int):
        return pfield
    if pfield[0] == 'p':
        return int(pfield[1:])
    if not pfieldNameToIndex:
        return 0
    return pfieldNameToIndex.get(pfield, 0)


def isAscii(s: str) -> bool:
    return all(ord(c)<128 for c in s)


from __future__ import annotations
import dataclasses
import emlib.misc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import *
    from . import engine
    import ipywidgets as ipy


def _guessStep(minval: float, maxval: float):
    diff = maxval - minval
    if diff <= 2:
        step = 0.001
    elif diff < 10:
        step = 0.01
    elif diff < 50:
        step = 0.05
    elif diff < 200:
        step = 0.1
    else:
        step = 1
    return step


def _guessRange(value: float, namehint: str = '') -> Tuple[float, float]:
    if namehint:
        if "freq" in namehint and 0 < value < 12000:
            return (0, 12000)
        if "amp" in namehint and 0 <= value <= 1:
            return (0, 1)
        if "cutoff" in namehint and 0 < value < 12000:
            return (0, 12000)
        if "midi" in namehint and 0 <= value <= 127:
            return (0, 127)
    if value < 0:
        val0, val1 = _guessRange(-value)
        return -val1, -val0
    if value < 1:
        return (0, 10)
    elif value < 50:
        return (0, 500)
    elif value < 100:
        return (0, 1000)
    else:
        return (0, value * 10)


@dataclasses.dataclass
class ParamSpec:
    descr: str
    minvalue: float
    maxvalue: float
    step: float = -1
    startvalue: Optional[float] = None
    widgetHint: str = 'slider'

    def __post_init__(self):
        if self.step == -1:
            self.step = _guessStep(self.minvalue, self.maxvalue)


def _stepToFormat(step: float) -> str:
    if step < 0.01:
        fmt = ".3f"
    elif step < 0.1:
        fmt = ".2f"
    elif step < 1:
        fmt = ".1f"
    else:
        fmt = "d"
    return fmt


def _jupySlider(name:str, startvalue: float, minvalue: float, maxvalue: float,
                callback:Callable, step:float=None, width='80%', log=False):
    import ipywidgets as ipy
    if step is None:
        step = _guessStep(minvalue, maxvalue)
    fmt = _stepToFormat(step)
    layout = ipy.Layout(width=width)
    if not log:
        s = ipy.FloatSlider(value=startvalue, min=minvalue, max=maxvalue,
                            step=step, description=name, layout=layout,
                            readout_format=fmt)
    else:

        s = ipy.FloatLogSlider(value=startvalue, min=minvalue, max=maxvalue,
                               step=step, description=name, layout=layout,
                               readout_format=fmt)
    if callback:
        s.observe(lambda change:callback(change['new']), names='value')
    return s


def _jupyEntry(name: str, startvalue: float, minvalue:float, maxvalue: float,
               callback: Callable):
    import ipywidgets as ipy
    step = 0.001
    w = ipy.BoundedFloatText(value=startvalue, min=minvalue, max=maxvalue,
                             step=step, description=name)
    if callback:
        w.observe(lambda change:print(change), names='value')
    return w


def interact(**sliders: Dict[str, Tuple[float, float, float, Callable]]):
    """
    Creates a set of interactive widgets

    Args:
        sliders: given as keywords. The key is used as the widget name. The
            value is a tuple (initialvalue, minvalue, maxvalue, callback)
            where callback has the form func(x) -> None and will be called
            with the current value of the widget

    Example
    =======

        from csoundengine import *
        from csoundengine.interact import *
        from IPython.display import display

        e = Engine()
        e.compile(r'''
        instr foo
          kamp = p4
          kmidi = p5
          asig vco2 lag:k(kamp, 0.1), lag:k(mtof:k(kmidi), 0.1)
          outch 1, asig
        endin
        ''')
        p1 = e.sched("foo", [0.1, 67])
        sliders = makeSliders(kamp=(e.getp(p1, 4, 0, 1, lambda x:e.setp(p1, 4, x),
                              kmidi=(e.getp(p1, 5, 0, 127, lambda x:e.setp(p1, 5, x))
        display(*sliders)
    """
    from IPython.display import display
    widgets = []
    for key, value in sliders.items():
        curvalue, minvalue, maxvalue, func = value
        s = _jupySlider(name=key, startvalue=curvalue, minvalue=minvalue,
                        maxvalue=maxvalue, callback=func)
        widgets.append(s)
    display(*widgets)


def interactPargs(engine: engine.Engine, p1: Union[float, str],
                  specs: Dict[Union[int, str], ParamSpec]=None,
                  **namedSpecs):
    """
    Interact with pfields of an event

    Depending on the context this will create a set of sliders to
    interact with the dynamic pfields of a running event

    Example
    =======

    .. code::

        from csoundengine import *
        from csoundengine.interact import *
        e = Engine()
        e.compile(r'''
        instr 10
          kamp = p4
          kmidi = p5
          a0 vco2 kamp, lag:k(mtof:k(kmidi), 0.1)
          outch 1, a0
        ''')
        eventid = e.sched(10, args=[0.1, 67])
        interactPfields(e, eventid,
                        specs={4: ParamSpec("kamp", 0, 1),
                               5: ParamSpec("kmidi", 0, 127)})
        # This is the same:
        interactPfields(e, eventid,
                        p4=ParamSpec("kamp", 0, 1),
                        p5=ParamSpec("kmidi", 0, 127))

    """
    allspecs = {}
    if specs:
        allspecs.update(specs)
    if namedSpecs:
        allspecs.update(namedSpecs)
    if emlib.misc.inside_jupyter():
        return _jupyInteractPargs(engine=engine, p1=p1, specs=allspecs)
    else:
        raise RuntimeError("interact is only supporte inside a jupyter session at the"
                           " moment.")


def _jupyInteractPargs(engine: engine.Engine, p1: Union[float, str],
                       specs: Dict[Union[int, str], ParamSpec]=None,
                       stopbutton=True,
                       width='80%'):
    """

    .. note::

        This function should only be called inside a jupyter session

    Example
    =======

    .. code::

        from csoundengine import *
        e = Engine()
        e.compile(r'''
        instr 10
          kamp = p4
          kmidi = p5
          a0 vco2 kamp, lag:k(mtof:k(kmidi), 0.1)
          outch 1, a0
        endin
        ''')
        event = e.sched(10, args=[0.1, 67])
        e.interact(event, p4=ParamSpec('kamp', 0, 1){4: ParamSpec('kamp', 0, 1)})
    """
    from IPython.display import display
    import ipywidgets as ipy
    widgets = []
    if stopbutton:
        button = ipy.Button(description="Stop")
        button.on_click(lambda *args, e=engine, p1=p1: e.unsched(p1))
        widgets.append(button)

    for key, spec in specs.items():
        idx = key if isinstance(key, int) else int(key[1:])
        value0 = engine.getp(p1,idx) if spec.startvalue is None else spec.startvalue
        if spec.widgetHint == 'slider':
            w = _jupySlider(spec.descr, startvalue=value0,
                            minvalue=spec.minvalue, maxvalue=spec.maxvalue,
                            width=width,
                            callback=lambda val, p1=p1, idx=idx: engine.setp(p1, idx, val))
        elif spec.widgetHint == 'entry':
            w = _jupyEntry(spec.descr, startvalue=value0,
                           minvalue=spec.minvalue, maxvalue=spec.maxvalue,
                           callback=lambda val, p1=p1, idx=idx: engine.setp(p1, idx, val))
        else:
            raise ValueError(f"Widget hint not understood: {spec.widgetHint}")
        widgets.append(w)
    display(*widgets)




from __future__ import annotations
import cachetools
from . import internalTools
from dataclasses import dataclass
import subprocess
import shutil
import re


@cachetools.cached(cache=cachetools.TTLCache(1, 20))
def isPipewireRunning() -> bool:
    return internalTools.isrunning("pipewire")


@cachetools.cached(cache=cachetools.TTLCache(1, 20))
def isPulseaudioRunning() -> bool:
    """
    Returns True if the pulseaudio server is running

    .. note:: pulseaudio can be running on pipewire even if this returns False
    """
    if not shutil.which("pulseaudio"):
        return False
    status = subprocess.call(["pulseaudio", "--check"])
    return status == 0


@dataclass
class PipewireInfo:
    sr: int
    quantum: int
    isPulseServer: bool


@dataclass
class PulseaudioInfo:
    sr: int
    numchannels: int
    onPipewire: bool = False


@cachetools.cached(cache=cachetools.TTLCache(1, 20))
def pulseaudioInfo() -> PulseaudioInfo:
    sr = 48000
    numchannels = 2
    onPipewire = False
    if not shutil.which("pactl"):
        raise RuntimeError("pactl not found")
    output = subprocess.check_output(['pactl', 'info']).decode('utf-8')
    for line in output.splitlines():
        line = line.strip()
        if line.startswith("Server Name"):
            serverName = line.split(":", maxsplit=1)[1]
            if "PipeWire" in serverName:
                onPipewire = True
        elif line.startswith("Default Sample Specification"):
            spec = line.split(":", maxsplit=1)[1]
            samplefmt, numchnlstr, srstr = spec.split()
            numchannels = int(numchnlstr[:-2])
            sr = int(srstr[:-2])
    return PulseaudioInfo(sr=sr, numchannels=numchannels, onPipewire=onPipewire)


@cachetools.cached(cache=cachetools.TTLCache(1, 20))
def pipewireInfo() -> PipewireInfo:
    assert shutil.which('pw-cli')
    sr = 48000
    quantum = 1024
    output = subprocess.check_output(['pw-cli', 'info', '0']).decode('utf-8')
    for line in output.splitlines():
        if match := re.search(r'default\.clock\.rate = "([0-9]+)"', line):
            sr = int(match.group(1))
        elif match := re.search(r'default\.clock\.quantum = "([0-9]+)"}', line):
            quantum = int(match.group(1))
    pulseinfo = pulseaudioInfo()
    return PipewireInfo(sr=sr, quantum=quantum, isPulseServer=pulseinfo.onPipewire)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

version = (1, 6, 4)

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='csoundengine',
    python_requires=">=3.9",
    version=".".join(map(str, version)),
    description='A synthesis framework using csound',
    long_description=long_description,
    author='Eduardo Moguillansky',
    author_email='eduardo.moguillansky@gmail.com',
    url='https://github.com/gesellkammer/csoundengine',
    packages=[
        'csoundengine',
    ],
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "JACK-client",
        "ctcsound",
        "sndfileio>=1.8.1",
        "emlib>=1.1",
        "cachetools",
        "configdict>=0.8",
        "bpf4",
        "numpyx>=0.5",
        "pitchtools>=1.1",
        "appdirs",
        "pygments",
        "sf2utils",
        "jupyter",
        "ipywidgets",
        "risset"
    ],
    license="BSD",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    package_data={'': ['data/*']},
    include_package_data=True,
    zip_safe=False
)



"""
This module defines some %magic / %%magic to work with an Engine or a Session

Magics defined
==============

%csound
-------

Syntax::

    %csound setengine <enginename>         : Sets the default Engine

%%csound
--------

Compile the code in this cell
"""
from __future__ import annotations
from IPython.core.getipython import get_ipython
from IPython.core.magic import Magics, magics_class, cell_magic, line_cell_magic
from IPython.display import display, HTML
from . import engine as _engine
from .config import logger, config
from . import csoundlib

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import *


def _splitonce(s: str):
    parts = s.split(maxsplit=1)
    return parts if len(parts) == 2 else (parts[0], '')


@magics_class
class EngineMagics(Magics):
    """Implement magic commands for Csound."""
    def __init__(self, shell):
        super().__init__(shell)
        self.currentEngine: Optional[_engine.Engine] = None
        self._lineCommands = {
            'setengine': self._cmd_setengine
        }

    def _resolveEngine(self, line: str) -> Optional[_engine.Engine]:
        if line:
            engineName = line.strip()
            engine = _engine.getEngine(engineName)
            if engine is None:
                logger.warning(f"Engine {engineName} not known. "
                               f"Possible engines: {_engine.activeEngines()}")
                return None
            return engine
        if self.currentEngine:
            return self.currentEngine
        engines = _engine.activeEngines()
        if not engines:
            print("No active Engine. After creating an Engine, "
                  "do `%csound <enginename>` to set it as the default Engine")
            return None
        return _engine.getEngine(list(engines)[-1])

    def _cmd_setengine(self, line:str) -> None:
        engine = _engine.getEngine(line)
        if not engine:
            logger.error(f"Engine {engine} unknown. Active engines: "
                         f"{_engine.activeEngines()}")
        else:
            self.currentEngine = engine

    @line_cell_magic
    def csound(self, line, cell=None):
        """
        %csound and %%csound magics.

        ``%csound setengine <enginename>``
            Set the default engine for any subsequent magic. Otherwise the last
            created Engine will be used

        ``%%csound [<enginename>]``
            Cell magic, compiles any code within the given engine, or the default
            engine if no engine is given.

        Example
        -------

        .. code::

            from csoundengine import *
            e = Engine("foo")

            # In another cell
            %%csound
            instr 10
                kamp = p4
                kmidi = p5
                a0 oscili kamp, mtof:k(kmidi)
                aenv linsegr 0, 0.01, 1, 0.01, 0
                a0 *= aenv
                outch 1, a0
            endin

            # Schedule an A4 note lasting 20 seconds
            eventid = e.sched(10, 0, 20, [0.1, 69])

            # Change the pitch
            e.pwrite(eventid, "kmidi", 67)
        """
        # line magic
        if cell is None:
            assert line
            cmd, rest = _splitonce(line)
            func = self._lineCommands.get(cmd)
            if not func:
                logger.error(f"command {cmd} unknown. Possible commands: "
                             f"{self._lineCommands.keys()}")
                return None
            func(rest)
            return None
        # cell magic
        engine = self._resolveEngine(line)
        if not engine:
            return None
        self.currentEngine = engine
        self.currentEngine.compile(cell, block=True)
        html = csoundlib.highlightCsoundOrc(cell)
        display(HTML(html))
        return None

    @cell_magic
    def definstr(self, line: str, cell: str) -> None:
        """
        Defines a new Instr inside the current Session

        Example
        =======

        .. code-block:: python

            from csoundengine import *
            e = Engine()

        .. code-block:: csound

            %%definstr foo
            |kamp=0.5, kfreq=1000|
            a0 oscili kamp, kfreq
            aenv linsegr 0, 0.01, 1, 0.01, 0
            a0 *= aenv
            outch 1, a0

        The last block is equivalent to::

            s = e.session()
            s.defInstr("foo", r'''
                |kamp=0.5, kfreq=1000|
                a0 oscili kamp, kfreq
                aenv linsegr 0, 0.01, 1, 0.01, 0
                a0 *= aenv
                outch 1, a0
            ''')
        """
        parts = line.split()
        if len(parts) == 2:
            enginename, instrname = parts
            self.currentEngine = _engine.getEngine(enginename)
        elif len(parts) == 1:
            instrname = parts[0]
            if not self.currentEngine:
                engines = _engine.activeEngines()
                if not engines:
                    print("No active Engine. After creating an Engine, "
                          "do `%csound <enginename>` to set it as the default Engine")
                    return None
                self.currentEngine = _engine.getEngine(list(engines)[-1])
        else:
            raise SyntaxError("Syntax: %%definstr [engine] instrname")
        session = self.currentEngine.session()
        instr = session.defInstr(name=instrname, body=cell)
        display(HTML(instr._repr_html_()))


def load_ipython_extension(ip):
    if config['magics_print_info']:
        print("csoundengine.magic extension loaded")
        print("Magics available: %csound, %%csound, %%definstr")
    ip.magics_manager.register(EngineMagics)
    ip.user_ns['Engine'] = _engine.Engine



from __future__ import annotations
from configdict import ConfigDict
from datetime import datetime
import emlib.dialogs
import os


__all__ = ('state')

_home = os.path.expanduser("~")

def _isofmt(t:datetime) -> str:
    """Returns the time in iso format"""
    return t.isoformat(':', 'minutes')


_defaultState = {
    'last_run': datetime(1900, 1, 1).isoformat(),
    'soundfont_last_dir': _home,
    'soundfile_last_dir': _home,
}


state = ConfigDict("csoundengine.state", _defaultState)


def openFile(key, filter="All (*.*)", title="Open File"):
    folder = state[key]
    f = emlib.dialogs.selectFile(directory=folder, filter=filter, title=title)
    if f:
        folder = os.path.split(f)[0]
        state[key] = folder
    return f


def saveFile(key, filter="All (*.*)", title="Save File"):
    folder = state[key]
    f = emlib.dialogs.saveDialog(directory=folder, filter=filter, title=title)
    if f:
        folder = os.path.split(f)[0]
        state[key] = folder
    return f


def openSoundfile(filter="Soundfiles (*.wav, *.flac, *.aif, *.aiff)",
                  title="Open Soundfile",
                  ensureSelection=False):
    out = openFile("soundfile_last_dir", filter=filter, title=title)
    if not out and ensureSelection:
        raise ValueError("No output selected")
    return out


def openSoundfont(filter="Soundfont (*.sf2)", title="Open Soundfont", ensureSelection=False):
    out = openFile("soundfont_last_dir", filter=filter, title=title)
    if not out and ensureSelection:
        raise ValueError("No soundfont selected")
    return out




